#include "updater.h"
#include "stdio.h"
#include "string.h"
#include "malloc.h"
#include "libgen.h"
#include "sys/stat.h"

#include "wzdb.h"

char my_version[255];
char update_url[1024];
int force_update = 0;
int run_mode = 0;

unsigned long prgwritten = 0;

static size_t write_data(void *ptr, size_t size, size_t nmemb, void *stream) {
	int written = fwrite(ptr, size, nmemb, (FILE *)stream);
	return written;
}

static size_t write_data_prg(void *ptr, size_t size, size_t nmemb, void *stream) {
	static int dp = 0;
	dp++; 
	if (dp==50) { printf("O\x08"); fflush(stdout); }
	if (dp==100) { printf("o\x08"); fflush(stdout); }
	if (dp==150) { printf(":\x08"); fflush(stdout); }
	if (dp==200) { printf("."); fflush(stdout); dp = 0; }
	int written = fwrite(ptr, size, nmemb, (FILE *)stream);
	prgwritten += written;
	return written;
}

static void print_formatted_size(float dw) {
	if (dw > 1024) {
		dw = dw / 1024.0f;
		if (dw > 1024) {
			dw = dw / 1024.0f;
			printf(". %.2fmB\n", dw);
		} else {
			printf(". %.2fkB\n", dw);
		}
	} else {
		printf(". %.2fb\n", dw);
	}
}

int http_get_file(char* url, char* dest) {
	// Create a curl request for the url of the latest version
	CURL* ch = curl_easy_init();
	FILE* ptr = fopen(dest, "w");
		curl_easy_setopt(ch, CURLOPT_NOPROGRESS, 1);
		curl_easy_setopt(ch, CURLOPT_URL, url);
		curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION, write_data);
		curl_easy_setopt(ch, CURLOPT_WRITEDATA, ptr);
	curl_easy_perform(ch);
	fclose(ptr);
	curl_easy_cleanup(ch);
	return 1;
}

int http_get_file_progress(char* url, char* dest) {
	// Create a curl request for the url of the latest version
	CURL* ch = curl_easy_init();
	prgwritten = 0;
	FILE* ptr = fopen(dest, "w");
		curl_easy_setopt(ch, CURLOPT_NOPROGRESS, 1);
		curl_easy_setopt(ch, CURLOPT_URL, url);
		curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION, write_data_prg);
		curl_easy_setopt(ch, CURLOPT_WRITEDATA, ptr);
	curl_easy_perform(ch);
	fclose(ptr);
	curl_easy_cleanup(ch);
	return 1;
}

void check_and_mkdir(char* filename) {
	if (strcmp(filename,".") == 0) return;
	char path[256];
	strcpy(path,filename);
	char* dir = dirname(path);
	// Go recursive
	check_and_mkdir(dir);
	mkdir(dir,0755);
}

int main(int argc, char** argv) {

	int i;
	
	memset(my_version,'\0',sizeof(my_version));
	memset(update_url,'\0',sizeof(update_url));
	force_update = 0;
	run_mode = 0;

	// Hardcoded, change this to what the initial option should be.
	strcpy(update_url,"http://files.noccy.com/wzupdate/\0");

	for (i = 1; i < argc; i++) {
		if (0 == strcmp("-h", argv[i])) {
			print_usage();
			return 2;
		}
		if (0 == strcmp("-v", argv[i])) {
			if (argc > i+1) {
				// TODO: Check if this really is a version string
				strcpy(my_version,argv[i+1]);
				i++;
			}
		} else 
		if (0 == strcmp("-i", argv[i])) {
			printf("Initial setup mode\n");
			strcpy(my_version,"0.0.0");
		} else
		if (0 == strcmp("-f", argv[i])) {
			printf("Forcing update\n");
			force_update = 1;
		} else
		if (0 == strcmp("-r", argv[i])) {
			printf("Will launch after update\n");
		} else
		if (0 == strcmp("-l", argv[i])) {
			if (argc > i+1) {
				strcpy(update_url,argv[i+1]);
				i++;
			}
		}
		
	}

	FILE* fh = fopen("wzupdate.ver","r");
	if (NULL != fh) {
		printf("Checking for game version: ");
		if (fscanf(fh,"%s",my_version)) {
			printf("%s\n", my_version);
		} else {
			if (0 == strlen(my_version)) {
				printf("I don't know my version. You need to launch me with the -v switch!\n");
				return 3;
			}
		}
	}
	

	char url[512];
	memset(url,0,sizeof(url));
	
	// Add LATEST to the end of the URL
	strcpy(url,update_url);
	strcat(url,"LATEST\0");
	
	// Grab the latest version
	char version_latest[64];
	if (http_get_file(url,"wzupdate.latest")) {
		// And read it back in
		printf("Looking for updates: ");
		FILE* ptr = fopen("wzupdate.latest", "r");
		if (fscanf(ptr,"%s",version_latest)) {
			printf("Latest is %s\n", version_latest);
		}
		fclose(ptr);
	}

	if ((force_update) || (1 == strcmp(version_latest,my_version))) {
		// Grab the update manifest/checksums
		printf("An update is available! Downloading manifest\n");
		strcpy(url,update_url);
		strcat(url,"wzupdate-");
		strcat(url,version_latest);
		strcat(url,".md5\0");

		// Match files against database
		if (http_get_file(url,"wzupdate.md5")) {
			char downloads[256][256];
			memset(downloads,0,sizeof(downloads));
			int downloadcount = 0;
			if (init_wz_db() && open_wz_db()) {
				FILE* ptr = fopen("wzupdate.md5", "r");
				printf("Checking files: ");
				int files_in_db = 0;
				int files_new = 0;
				int files_updated = 0;
				char checksum[32];
				char filename[256];
				while (fscanf(ptr,"%s %s", checksum, filename) != EOF) {
					int state = entry_match_md5(filename, checksum);
					switch(state) {
						case 0:
							files_in_db++;
							break;
						case 1:
							strcpy(downloads[downloadcount++],filename);
							files_in_db++; 
							files_updated++; 
							break;
						case 2:
							strcpy(downloads[downloadcount++],filename);
							files_new++;
							break;
					}
				}
				fclose(ptr);
				printf("%d in database, %d new, %d needs updating.\n", files_in_db, files_new, files_updated);
				int i;
				for(i = 0; i < downloadcount; i++) {
					check_and_mkdir(downloads[i]);
					printf("Downloading %s [%2d/%2d] ", downloads[i], i + 1, downloadcount);
					strcpy(url,update_url);
					strcat(url,version_latest);
					strcat(url,"/");
					strcat(url,downloads[i]);
					strcat(url,"\0");
					if (http_get_file_progress(url,downloads[i])) {
						print_formatted_size((float)prgwritten);
					} else {
						printf("! Error\n");
					}
				}
				close_wz_db();
				FILE* fh = fopen("wzupdate.ver","w");
				fprintf(fh, "%s",version_latest);
				fclose(fh);
			} else {
				printf("Error opening database!\n");
				return 5;
			}
		}
	} else {
		printf("No update neccesary.\n");
	}
	
	// Launch the game if requested
	if (run_mode) {
		// Load game here.
	}
	
	return 0;

}

void print_usage(void) {
	printf("wzupdater - Warzone2100 Update Utility by Noccy\n");
	printf("(c) 2010, Noccy Labs. Released under the GNU GPL v3.\n");
	printf("To everyone at the wz2100.net forums :)\n\n");
	
	printf("Usage:\n");
	printf("  wzupdater -h\n");
	printf("  wzupdater [-u] [-f] [-v MYVERSION] [-l UPDATEURL]\n");
	printf("  wzupdater -i\n");
	printf("\n");
	printf("Options:\n");
	printf("  -h    Show this help\n");
	printf("  -v    Set local version\n");
	printf("  -f    Force update to latest files\n");
	printf("  -l    Override update URL\n");
	printf("  -f    Force update to latest files\n");
	printf("  -i    Initial setup mode, will update what is needed based on hashes\n");
	printf("  -r    Run mode -- Update, then launch binary\n");
	printf("\n");
	printf("Examples:\n");
	printf("  wzupdater -u  -- Update to latest version based on local database\n");
	printf("  wzupdater -u -v 2.3.4 -f  -- Update local as if 2.3.4 to latest version\n");
	printf("\n");
	printf("Make sure to run wzupdater from the wz2100 folder\n");
}
